"""
Module containing the Command-Line Interface control logic.
"""
import argparse
import logging
from pathlib import Path
import sys

from physiofit.base.io import IoHandler, StandardDevs, ConfigParser

def args_parse():
    """
    Parse arguments from user input.
    :return: Argument Parser object
    :rtype: class: argparse.ArgumentParser
    """

    parser = argparse.ArgumentParser(
        "Physiofit: Extracellular flux estimation software"
    )

    # Parse data arguments (tsv + json)
    parser.add_argument(
        "-d", "--data", type=str,
        help="Path to data file in tabulated format (txt or tsv)"
    )
    parser.add_argument(
        "-c", "--config", type=str,
        help="Path to config file in yaml format"
    )
    parser.add_argument(
        "-m", "--model", type=str,
        help="Which model should be chosen. Useful only if generating related config file"
    )

    # Parse developer arguments
    parser.add_argument(
        "-g", "--galaxy", action="store_true",
        help="Is the CLI being used on the galaxy platform"
    )
    parser.add_argument(
        "--list", action="store_true",
        help="Return the list of models in model folder"
    )
    parser.add_argument(
        "-v", "--debug_mode", action="store_true",
        help="Activate the debug logs"
    )

    # Parse selective output path arguments (for galaxy implementation mostly)
    parser.add_argument(
        "-op", "--output_pdf", type=str,
        help="Path to output the pdf file containing plots"
    )
    parser.add_argument(
        "-of", "--output_fluxes", type=str,
        help="Path to output the flux results"
    )
    parser.add_argument(
        "-os", "--output_stats", type=str,
        help="Path to output the khi² test"
    )
    parser.add_argument(
        "-oc", "--output_config", type=str,
        help="Path to output the yaml config file"
    )

    return parser

def run(data, args, logger, exp=None):

    io = IoHandler()
    logger.info(f"Input Data: \n{io.data}")
    io.home_path = Path(args.data).parent
    logger.info(f"Home path: {io.home_path}")
    logger.info(f"Reading configuration file at {args.config}")
    io.configparser = io.read_yaml(args.config)
    logger.info(f"Config File:\n{io.configparser}")
    model = io.select_model(io.configparser.model["model_name"], data)
    model.get_params()
    logger.info(f"Selected Model:\n{model}")
    model = io.configparser.update_model(model)
    logger.info(f"Updated Model: \n{model}")
    logger.info(f"Model Data: \n{model.data}")
    fitter = io.initialize_fitter(
        model.data,
        model=model,
        mc=io.configparser.mc,
        iterations=io.configparser.iterations,
        sd=io.configparser.sds,
        debug_mode=args.debug_mode
    )
    fitter.optimize()
    if fitter.mc:
        fitter.monte_carlo_analysis()
    fitter.khi2_test()
    if exp is not None:
        res_path = io.home_path / (io.home_path.name + "_res") / exp
    else:
        res_path = io.home_path / (io.home_path.name + "_res")
    if not res_path.is_dir():
        res_path.mkdir(parents=True)
    io.output_report(fitter, res_path)
    io.output_plots(fitter, res_path)
    io.output_pdf(fitter, res_path)
    io.figures = []

def generate_config(args, data, logger):

    logger.info("Launching in configuration file generator mode")
    # Run checks
    if args.model is None:
        # Must be path to model.py file or name of a model
        raise ValueError(
            f"Please select a model to generate the associated configuration file"
        )
    if args.output_config is None:
        raise ValueError(
            "Please select an output directory for the configuration file"
        )
    try:
        io = IoHandler()
        if Path(args.model).is_file():
            model_class = io.read_model(args.model)
            model = model_class(data)
            model.get_params()

        else:
            model = io.select_model(args.model, data)
            model.get_params()
    except Exception:
        logger.error("There was an error while initialising the model")
        raise

    # Build config parser and export configuration file
    default_sds = StandardDevs({col: 0.2 for col in model.name_vector})
    configparser = ConfigParser(
        path_to_data=args.data,
        selected_model=model,
        sds=default_sds,
        mc=True,
        iterations=100
    )
    configparser.export_config(args.output_config)
    logger.info(f"Finished exporting the configuration file to {args.output_config}")
    sys.exit()

def process(args):

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    out_stream = logging.StreamHandler()
    formatter = logging.Formatter()
    out_stream.setFormatter(formatter)
    if args.debug_mode:
        out_stream.setLevel(logging.DEBUG)
    else:
        out_stream.setLevel(logging.INFO)
    logger.addHandler(out_stream)

    if args.list:
        IoHandler.get_model_list()
        sys.exit()

    if args.data is None:
        raise ValueError("A path towards the data must be given")

    # Ensure that the path to data is correct
    path_to_data = Path(args.data)
    if not path_to_data.is_file():
        raise ValueError(
            f"The data path is not correct. Please check and try again. Input data path: \n{args.data}"
        )
    # Ensure that the input file is a tsv
    if not path_to_data.suffix in [".tsv", ".txt"]:
        raise TypeError(
            f"The input data must be in tsv/txt format. Detected format: {path_to_data.suffix}"
        )

    # Read & check data
    data = IoHandler.read_data(str(path_to_data))

    # If no configuration file is present we assume the user wants to generate a default one
    if args.config is None:
        generate_config(args, data, logger)

    # If configuration file is present we launch the optimization
    if "experiments" in data.columns:
        experiments = list(data["experiments"].unique())
        data = data.set_index("experiments")
        for exp in experiments:
            logger.info(f"Running optimization for {exp}")
            run_data = data.loc[exp, :].sort_values("time").copy()
            run(run_data, args, logger, exp)
    else:
        logger.info("Running optimization")
        run_data = data.sort_values("time")
        run(run_data, args, logger)
    logger.info("Done!")
    sys.exit()

def main():
    parser = args_parse()
    args = parser.parse_args()
    process(args)



from subprocess import run
import sys
from pathlib import Path

from threading import Thread
import requests

import physiofit
import physiofit.ui.cli


def get_last_version():
    """Get last Physiofit version."""
    try:
        pf_path = Path(physiofit.__file__).parent
        # Get the version from pypi
        response = requests.get(f'https://pypi.org/pypi/physiofit/json')
        latest_version = response.json()['info']['version']
        with open(str(Path(pf_path, "last_version.txt")), "w") as f:
            f.write(latest_version)
    except Exception:
        pass

def main():
    """The main routine"""

    if len(sys.argv) > 1:
        physiofit.ui.cli.main()
    else:
        thread = Thread(target=get_last_version)
        thread.start()
        path_to_app = Path(physiofit.__file__).parent
        path_to_app = path_to_app / "ui/gui.py"
        run(["streamlit", "run", str(path_to_app)])

if __name__ == "__main__":
    sys.exit(main())


"""Module to handle inputs and outputs for PhysioFit"""

from __future__ import annotations
import importlib
import logging
import os
import sys
from pathlib import Path
from io import BytesIO

import matplotlib.pyplot as plt
import numpy as np

# Switch matplotlib logger to higher level to not get debug logs in root logger
logging.getLogger("matplotlib").setLevel(logging.WARNING)

from matplotlib.backends.backend_pdf import PdfPages
from pandas import DataFrame, read_csv, concat
import yaml

from physiofit import __file__
from physiofit.base.fitter import PhysioFitter
from physiofit.models.base_model import StandardDevs, Bounds

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class IoHandler:
    """
    Input/Output class that handles the former and initializes the
    PhysioFitter component object. It is the preferred interface for
    interacting with the PhysioFit package.

    """

    allowed_keys = {
        "sd", "model", "iterations", "mc", "debug_mode"
    }

    def __init__(self):

        self._output_type = []
        self.figures = []
        self.models = []
        self.fitter = None
        self.output = None
        self.data = None
        self.names = None
        self.simulated_data_sds = None
        self.simulated_data = None
        self.experimental_data = None
        self.home_path = None
        self.data_path = None
        self.res_path = None
        self.has_config_been_read = False
        self.experiments = None
        self.multiple_experiments = False

    @staticmethod
    def read_data(data: str) -> DataFrame:
        """
        Read initial data file (csv or tsv)

        :param data: str containing the relative or
                             absolute path to the data
        :return: pandas DataFrame containing the data
        """
        try:
            if isinstance(data, str):
                data_path = Path(data).resolve()
            # .dat file type for galaxy implementation
                if data_path.suffix in [".txt", ".tsv", ".dat"]:
                    data = read_csv(str(data_path), sep="\t")
                elif data_path.suffix == ".csv":
                    data = read_csv(str(data_path), sep=";")
                else:
                    if not data_path.exists():
                        raise ValueError(f"{data_path} is not a valid file")
                    else:
                        raise TypeError(
                            f"{data_path} is not a valid format. "
                            f"Accepted formats are .csv, .txt or .tsv"
                        )
            elif issubclass(type(data), BytesIO):
                data = read_csv(data, sep="\t")
            else:
                raise TypeError(f"Input data file is not of right type. Accepted types: file-like (bytes) or string")
        except Exception:
            logger.exception("There was an error while reading the data")
            raise IOError(
                "Error while reading data. Please ensure you have the right file format (txt, tsv or bytes)"
            )

        IoHandler._verify_data(data)
        return data

    def select_model(self, model_name, data=None):
        """
        Select a model from the list of models in the model folder of the package src directory
        """

        self.get_models(data)
        for x in self.models:
            if x.model_name == model_name:
                model = x
        return model

    def read_model(self, model_file):
        """
        Import and return the model class from .py file containing the model.

        .. warning: ONLY USE THIS FUNCTION ON TRUSTED PYTHON FILES. Reading code from untrusted sources can lead to
                    propagation of viruses and compromised security.

        :param model_file: path to the model.py file to import
        """

        if not Path(model_file).is_file() or Path(model_file).suffix != ".py":
            raise ValueError(
                "The given path is not valid. The path must point to a .py file containing the module "
                "from which the model will be loaded."
            )

        spec = importlib.util.spec_from_file_location("module_to_import", fr"{model_file}")
        module = importlib.util.module_from_spec(spec)
        sys.modules["module_to_import"] = module
        spec.loader.exec_module(module)
        model_class = getattr(module, "ChildModel")

        return model_class


    @staticmethod
    def _verify_data(data: DataFrame):
        """
        Perform checks on DataFrame returned by the _read_data function

        :param data: pandas DataFrame containing the data
        :return: None
        """

        if not isinstance(data, DataFrame):
            raise TypeError(
                "There was an error reading the data: "
                "DataFrame has not been generated"
            )

        for x in ["time", "X"]:
            if x not in data.columns:
                raise ValueError(f"Column {x} is missing from the dataset")

        if "experiment" in data.columns:
            if len(data.columns) <= 3:
                raise ValueError(f"Data does not contain any metabolite columns")
        else:
            if len(data.columns) <= 2:
                raise ValueError(f"Data does not contain any metabolite columns")

        for x in data.columns:
            if x != "experiments" and data[x].dtypes != np.int64 and data[x].dtypes != np.float64:
                raise ValueError(
                    f"Column {x} has values that are not of numeric type"
                )
            if all(data[x].isnull()) or all(data[x].isna()):
                raise ValueError(
                f"The column {x} contains only null or NA values"
            )

    @staticmethod
    def get_model_list():
        model_dir = Path(__file__).parent / "models"
        # Make fake dataframe
        df = DataFrame(
            columns=["time", "X"],
        )
        for file in os.listdir(str(model_dir)):
            if "model_" in file:
                module = importlib.import_module(
                    f"physiofit.models.{file[:-3]}"
                )
                model_class = getattr(module, "ChildModel")
                model = model_class(df)
                print(model.model_name)
        return
        
    def get_models(self, data=None):
        """
        Read modules containing the different models and add them to models attribute

        :return: list containing the different model objects
        """

        model_dir = Path(__file__).parent / "models"
        for file in os.listdir(str(model_dir)):
            if "model_" in file:
                module = importlib.import_module(
                    f"physiofit.models.{file[:-3]}"
                )
                model_class = getattr(module, "ChildModel")
                if data is not None:
                    self.models.append(model_class(data))
                else:
                    self.models.append(model_class(self.data))

    @staticmethod
    def read_yaml(yaml_file: str | bytes) -> ConfigParser:

        """
        Import raml configuration file and parse keyword arguments

        :param yaml_file: path to the yaml file or json file
        :return config_parser: Dictionary containing arguments parsed from yaml file
        """

        # Load config file
        try:
            if isinstance(yaml_file, str) or issubclass(type(yaml_file), BytesIO):
                config_parser = ConfigParser.from_file(yaml_file)
            else:
                raise TypeError(
                    f"Trying to read object that is not a file or path to file: {yaml_file}"
                )
        except Exception as e:
            raise IOError(
                f"Error while reading yaml configuration file {yaml_file}. "
                f"\nTraceback:\n\n{e}"
                )
        return config_parser

    def initialize_fitter(self, data, **kwargs):
        """
        Initialize a PhysioFitter object

        :param data: input data
        :param kwargs: Keyword arguments for fitter initialization
        :return: None
        """
        #
        # wrong_keys = []

        # Initialize fitter
        fitter = PhysioFitter(
            data=data,
            model=kwargs["model"],
            mc=kwargs["mc"] if "mc" in kwargs else True,
            iterations=kwargs["iterations"] if "iterations" in kwargs else 100,
            sd=kwargs["sd"] if "sd" in kwargs else StandardDevs(),
            debug_mode=kwargs["debug_mode"] if "debug_mode" in kwargs else False
        )

        if "sd" not in kwargs:
            fitter.sd.update(
                {"X" : 0.2}
            )
            for col in self.data.columns[2:]:
                fitter.sd.update(
                    {col : 0.2}
                )

        fitter.initialize_sd_matrix()
        fitter.verify_attrs()
        logger.debug(
            f"Fitter attribute dictionary:\n{fitter.__dict__}"
        )

        return fitter


    def output_pdf(self, fitter, export_path: str | Path = None):
        """
        Handle the creation and output of a pdf file containing fit results as
        a plot

        :param export_path: Path to exported pdf. In local mode, it is sent to
                            the _res directory
        :return: None
        """

        if not self.figures:
            self.plot_data(fitter)

        try:
            with PdfPages(rf"{export_path}\plots.pdf") as pdf:
                for fig in self.figures:
                    pdf.savefig(fig[1])
        except Exception as e:
            raise IOError(f"Error while generating pdf:\n{e}")

    def output_plots(self, fitter, export_path):
        """
        Handle the creation and export of the different plots in svg format
        :return: None
        """

        if not self.figures:
            self.plot_data(fitter)

        try:
            for fig in self.figures:
                fig[1].savefig(rf"{export_path}\{fig[0]}.svg")
        except Exception:
            raise RuntimeError("Unknown error while generating output")

    def output_recap(self, export_path: str):

        if not isinstance(self.multiple_experiments, list):
            raise TypeError(
                "The multiple experiments attribute must be a list"
            )
        if not self.multiple_experiments:
            raise ValueError(
                f"It seems that the multiple experiments list is empty: {self.multiple_experiments}"
            )
        for idx, element in enumerate(self.multiple_experiments):
            if not isinstance(element, DataFrame):
                raise TypeError(
                    f"All the elements of multiple_experiments must be DataFrames. Wrong element type"
                    f"detected at indice {idx}"
                )
        final_df = concat(self.multiple_experiments)
        final_df = final_df.reset_index()
        final_df[["experiments", "parameter name"]] = final_df["index"].str.split(" ", expand=True)
        final_df.set_index(["experiments", "parameter name"], inplace=True)
        final_df.drop("index", axis=1, inplace=True)
        final_df.to_csv(f"{str(Path(export_path))}/summary.csv")


    def output_report(self, fitter, export_path: str |list = None):
        """
        Handle creation and export of the report containing stats from monte
        carlo analysis of optimization parameters

        :param export_paths: list of paths to export the stats and fluxes. [0]
                             is for stats and [1] for fluxes.
        """

        if isinstance(export_path, list):
            if len(export_path) != 2:
                raise ValueError(
                    f"Expected only 2 export paths and {len(export_path)}"
                    f" were detected"
                )
            stat_path = export_path[0]
            flux_path = export_path[1]
        else:
            flux_path = fr"{export_path}\flux_results.tsv"
            stat_path = fr"{export_path}\stat_results.tsv"

        logger.debug(
            f"Parameter Stats:\n{fitter.parameter_stats}"
        )

        # Get optimization results as dataframe
        opt_data = DataFrame.from_dict(fitter.parameter_stats,
                                       orient="columns")

        # Use IDs to clarify which parameter is described on each line
        opt_data.index = [param for param in fitter.model.parameters_to_estimate.keys()]
        opt_data.to_csv(flux_path, sep="\t")

        if isinstance(fitter.khi2_res, DataFrame):
            with open(stat_path, "w+") as stat_out:
                stat_out.write("==================\n"
                               "Khi² test results\n"
                               "==================\n\n")
                stat_out.write(
                    fitter.khi2_res.to_string(
                        header=False, justify="center"
                    )
                )
                if fitter.khi2_res.at["p_val", "Values"] < 0.95:
                    stat_out.write(
                        f"\n\nAt level of 95% confidence, the model fits the "
                        f"data good enough with respect to the provided "
                        f"measurement SD. Value: "
                        f"{fitter.khi2_res.at['p_val', 'Values']}"
                    )

                else:
                    stat_out.write(
                        f"\n\nAt level of 95% confidence, the model does not "
                        f"fit the data good enough with respect to the "
                        f"provided measurement SD. Value: "
                        f"{fitter.khi2_res.at['p_val', 'Values']}\n"
                    )

    def _get_plot_data(self, fitter):
        """
        Prepare data for plotting
        """

        # Initialize vectors and data for plotting
        if fitter.model.time_vector is not None:
            x = fitter.model.time_vector
        else:
            raise ValueError(
                "PhysioFitter model time_vector has not been initialized. "
                "Have you loaded in the data correctly?"
            )
        if fitter.experimental_matrix is not None:
            exp_mat = fitter.experimental_matrix
        else:
            raise ValueError(
                "PhysioFitter object does not have experimental data loaded in"
            )
        if fitter.simulated_matrix is not None:
            sim_mat = fitter.simulated_matrix
        else:
            raise ValueError("PhysioFitter simulated data does not exist yet")
        if fitter.matrices_ci is not None:
            sim_mat_ci = fitter.matrices_ci
            self.simulated_data_low_ci = DataFrame(
                columns=fitter.model.name_vector,
                index=x,
                data=sim_mat_ci["lower_ci"]
            )
            self.simulated_data_high_ci = DataFrame(
                columns=fitter.model.name_vector,
                index=x,
                data=sim_mat_ci["higher_ci"]
            )
        else:
            logger.warning(
                "Monte Carlo analysis has not been done, "
                "confidence intervals will not be computed"
            )

        self.experimental_data = DataFrame(
            columns=fitter.model.name_vector,
            index=x,
            data=exp_mat
        ).sort_index(level=0)
        self.simulated_data = DataFrame(
            columns=fitter.model.name_vector,
            index=x,
            data=sim_mat
        ).sort_index(level=0)

    def plot_data(self, fitter, display: bool = False):
        """
        Plot the data

        :param display: should plots be displayed
        """

        self._get_plot_data(fitter)
        self._draw_plots(fitter, display)

    def _draw_plots(self, fitter, display: bool):
        """
        Draw plots and assign them to the _figures attribute for later access
        in pdf and plot generation functions

        :param display: Should plots be displayed or not on creation
        """

        for element in fitter.model.name_vector:

            # Initialize figure object
            fig, ax = plt.subplots()
            fig.set_size_inches(9, 5)

            # Draw experimental points onto the Axe
            exp = ax.scatter(
                self.experimental_data.index,
                self.experimental_data[element],
                marker='o',
                color="orange"
            )
            exp.set_label(f"Exp {element}")

            # Draw the simulated line onto the Axe
            sim_line, = ax.plot(
                self.simulated_data.index,
                self.simulated_data[element],
                linestyle='--'
            )
            sim_line.set_label(f"Sim {element}")

            # If monte carlo analysis has been done add the
            # corresponding sd to the plot as a red area
            if fitter.matrices_ci is not None:
                ax = self._add_sd_area(element, ax)

            # Finishing touches
            ax.set(xlim=0, ylim=0)
            ax.set_xlabel("Time", fontsize=21)
            ax.set_ylabel("Concentration", fontsize=21)
            ax.legend(prop={"size": 18})
            ax.set_title(f"{element}", fontsize=23)
            ax.tick_params(axis='both', which='major', labelsize=18)
            fig.tight_layout()

            # Add the figure with the metabolite name as index
            # to the _figures attribute for later use
            self.figures.append((element, fig))

        if display:
            for _ in self.figures:
                plt.show()
        plt.close()

    def _add_sd_area(self, element: str, ax: plt.Axes):
        """
        Draw red area around the fitting line to show confidence intervals

        :param element: Which variable is being plotted
        :param ax: axes on which to draw the area before
                   returning to mother function
        """

        y1 = self.simulated_data_low_ci[element].to_numpy()
        y2 = self.simulated_data_high_ci[element].to_numpy()
        x = self.simulated_data.index
        ax.fill_between(x, y1, y2, alpha=.3, linewidth=0, color="red")
        return ax


class ConfigParser:

    allowed_keys = ["model", "sds", "mc", "iterations"]

    def __init__(
            self,
            selected_model,
            sds,
            mc,
            iterations,
            path_to_data=None
    ):

        self.path_to_data = path_to_data
        self.model = selected_model
        self.sds = StandardDevs(sds)
        self.mc = mc
        self.iterations = iterations

        if not isinstance(self.mc, bool):
            raise TypeError(
                f"The MonteCarlo option must be given as a boolean (True or False). Detected input: {self.mc}, "
                f"type: {type(self.mc)}"
            )
        if not isinstance(self.iterations, int):
            raise TypeError(
                f"Number of iterations must be an integer: Detected input: {self.mc}, type: {type(self.iterations)}"
            )


    @classmethod
    def from_file(cls, yaml_file):

        if isinstance(yaml_file, str):
            with open(yaml_file, 'r') as file:
                data = yaml.safe_load(file)
        else:
            data = yaml.safe_load(yaml_file)
        data_keys = [key for key in data.keys()]
        for key in cls.allowed_keys:
            if key not in data_keys:
                raise ValueError(
                    f"The key {key} is missing from the input config file"
                )

        try:
            return ConfigParser(
                path_to_data=data["path_to_data"],
                selected_model=data["model"],
                sds = data["sds"],
                mc = data["mc"],
                iterations = data["iterations"]
            )
        except KeyError:
            return ConfigParser(
                selected_model=data["model"],
                sds=data["sds"],
                mc=data["mc"],
                iterations=data["iterations"]
            )

    @classmethod
    def from_galaxy(cls, galaxy_yaml):
        pass

    def get_kwargs(self):
        return  {
            "path_to_data" : self.path_to_data,
            "model" : self.model,
            "mc" : self.mc,
            "iterations" : self.iterations,
            "sd" : self.sds
            }

    def update_model(self, model):

        if self.model["parameters_to_estimate"] is not None:
            model.parameters_to_estimate.update(self.model["parameters_to_estimate"])
        if self.model["bounds"] is not None:
            model.bounds.update(Bounds(self.model["bounds"]))
        return model

    def export_config(self, export_path):

        with open(fr"{export_path}/config_file.yml", "w") as file:
            data = {
                "model" : {
                    "model_name" : self.model.model_name,
                    "parameters_to_estimate" : self.model.parameters_to_estimate,
                    "bounds" : {name : f"{bounds[0], bounds[1]}" for name, bounds in self.model.bounds.items()}
                },
                "sds" : dict(self.sds),
                "mc" : self.mc,
                "iterations" : self.iterations,
                "path_to_data" : str(self.path_to_data)
            }
            yaml.safe_dump(
                data,
                file
            )


if __name__ == "__main__":
    import pandas as pd

    io_handler = IoHandler()
    data_file = pd.read_csv(
        r"C:\Users\legregam\Documents\Projets\PhysioFit\data\KEIO_test_data"
        r"\KEIO_ROBOT6_1\KEIO_ROBOT6_1.tsv",
        sep="\t"
    )
    io_handler.data = data_file
    io_handler.data = io_handler.data.sort_values(
        "time",
        ignore_index=True
    )
    io_handler.get_models()
    try:
        sd = {"X": 0.}
        for col in io_handler.data.columns[2:]:
            sd.update({col: 0.5})
    except Exception:
        raise
    io_handler.names = io_handler.data.columns[1:].to_list()
    model = io_handler.models[-1]
    model.get_params()
    print(model)
    keywargs = {
        "sd": sd,
        "model": model,
        "mc": True,
        "iterations": 100,
        "debug_mode": True,
    }
    io_handler.res_path = Path(
        r"C:\Users\legregam\Documents\Projets\PhysioFit\data\KEIO_test_data"
        r"\KEIO_ROBOT6_1")
    io_handler.initialize_fitter(kwargs=keywargs)
    io_handler.fitter.optimize()



from ast import literal_eval
import tkinter as tk
from tkinter import filedialog
from pathlib import Path
from copy import copy

import pandas as pd
import streamlit as st

import physiofit
from physiofit.base.io import IoHandler, ConfigParser
from physiofit.models.base_model import StandardDevs


class App:
    """
    Physiofit Graphical User Interface
    """

    def __init__(self):

        self.defaults = {
            "iterations" : 100,
            "sd" : StandardDevs(),
            "mc" : True
        }
        self.select_menu = None
        self.io = None
        self.update_info = None
        self.config_parser = None

    def start_app(self):
        """Launch the application"""

        st.set_page_config(page_title=f"PhysioFit (v{physiofit.__version__})")
        st.title(f"Welcome to PhysioFit (v{physiofit.__version__})")
        self.update_info = st.empty()
        self.check_uptodate()
        self.select_menu = st.selectbox(
            "Select a task to execute",
            ["Calculate extracellular fluxes",
             "Calculate degradation constant"]
        )
        if self.select_menu == "Calculate extracellular fluxes":
            self.io = IoHandler()
            self._build_flux_menu()
        else:
            st.header("Implementation in progress...")

    def check_uptodate(self):
        """Compare installed and most recent Physiofit versions."""
        try:
            pf_path = Path(physiofit.__file__).parent
            with open(str(Path(pf_path, "last_version.txt")), "r") as f:
                lastversion = f.read()
            if lastversion != physiofit.__version__:
                # change the next line to streamlit
                self.update_info = st.info(
                    f'New version available ({lastversion}). '
                    f'You can update PhysioFit with: "pip install --upgrade '
                    f'physiofit". Check the documentation for more '
                    f'information.'
                )
        except Exception:
            pass

    def _build_flux_menu(self):
        """Build the starting menu with the data upload button"""

        self.data_file = st.file_uploader(
            "Load a data file or a json configuration file",
            key="data_uploader",
            accept_multiple_files=False
        )

        if self.data_file:
            self._initialize_opt_menu()

    def _initialize_opt_menu(self):
        """
        Initialize all the options. If a yaml file is given as input, options
        are parsed from it. If the input is tsv, the defaults are used instead.
        """

        # Check file extension and initialize defaults
        file_extension = self.data_file.name.split(".")[1]
        if file_extension in ["yaml", "yml"]:
            try:
                # Get parameters from yaml file
                self.config_parser = self.io.read_yaml(self.data_file)
                # Load data into io_handler
                self.io.data = self.io.read_data(self.config_parser.path_to_data)
            except Exception:
                st.error("An error has occurred when reading the yaml configuration file.")
                raise
        elif file_extension  in ["tsv", "txt"]:
            try:
                self.io.data = self.io.read_data(self.data_file)
                # Initialize default SDs
                self.defaults["sd"].update({
                    "X": 0.5
                })
                self.defaults["sd"].update({
                    col: 0.2 for col in self.io.data.columns[2:]
                })
            except Exception:
                st.error("An error has occurred when reading the data.")
                raise
        else:
            raise KeyError(
                f"Wrong input file format. Accepted formats are tsv for data "
                f"files or json for configuration files. Detected file: "
                f"{self.data_file.name}")

        if "experiments" not in self.io.data.columns:
            raise ValueError(
                "'experiments' column missing from dataset"
            )
        self.io.data = self.io.data.sort_values(
            ["experiments", "time"], ignore_index=True
        )

        try:
            # Initialize the list of available models
            self.io.get_models()
        except Exception:
            st.error(f"An error has occurred when listing models from the models folder: "
                     f"\n{Path(__file__).parent / 'models'}. Please correct the model or submit an issue at "
                     f"github.com/MetaSys-LISBP/PhysioFit/issues")
            raise

        # Build menu
        submitted = self._initialize_opt_menu_widgets(
            file_extension
        )

        if submitted:
            try:
                self._get_data_from_session_state()
            except Exception:
                st.error("An error has occurred when initializing the model")
                raise
            if not self.io.home_path:
                raise ValueError("No output directory selected")
            self.config_parser = ConfigParser(
                path_to_data = self.io.home_path / self.data_file.name,
                selected_model= self.model,
                sds = self.sd,
                mc = self.mc,
                iterations = self.iterations
            )

            full_dataframe = self.io.data.copy()
            results_path = copy(self.io.res_path)
            experiments = list(self.io.data["experiments"].unique())
            self.io.multiple_experiments = []
            for experiment in experiments:
                with st.spinner(f"Running optimization for {experiment}"):
                    # final_table_dict = {}
                    self.model.data = full_dataframe[
                        full_dataframe["experiments"] == experiment
                    ].drop("experiments", axis=1).copy()

                    self.io.res_path = results_path / experiment
                    if not self.io.res_path.is_dir():
                        self.io.res_path.mkdir()
                    # Initialize the fitter object
                    self.io.names = self.io.data.columns[1:].to_list()
                    kwargs = self._build_fitter_kwargs()
                    fitter = self.io.initialize_fitter(
                        self.model.data,
                        model=kwargs["model"],
                        mc=kwargs["mc"],
                        iterations=kwargs["iterations"],
                        sd=kwargs["sd"],
                        debug_mode=kwargs["debug_mode"]
                    )
                    # Do the work
                    fitter.optimize()
                    if self.mc:
                        fitter.monte_carlo_analysis()
                    fitter.khi2_test()
                    df = pd.DataFrame.from_dict(
                        fitter.parameter_stats,
                        orient="columns"
                    )
                    df.index = [
                        f"{experiment} {param}" for param in fitter.model.parameters_to_estimate.keys()
                    ]
                    st.write(df)
                    self.io.multiple_experiments.append(df)

                    # Export results
                    self.io.output_report(fitter, self.io.res_path)
                    self.io.plot_data(fitter)
                    self.io.output_plots(fitter, self.io.res_path)
                    with st.expander(f"{experiment} plots"):
                        for fig in self.io.figures:
                            st.pyplot(fig[1])
                    self.io.output_pdf(fitter, self.io.res_path)
                    # Reset figures to free memory
                    self.io.figures = []
                    self.config_parser.export_config(self.io.res_path)
            self.io.data = full_dataframe
            self.io.res_path = results_path
            self.io.output_recap(results_path)

    def silent_sim(self):

        self.model.simulate(
            [param for param in self.model.parameters_to_estimate.values()],
            self.model.experimental_matrix,
            self.model.time_vector,
            self.model.fixed_parameters
        )

    def _initialize_opt_menu_widgets(self, file_extension):

        # Get model names and give as options to user
        model_options = [
            model.model_name for model in self.io.models
        ]
        model_options.insert(0, "--")
        model_name = st.selectbox(
            label="Model",
            options=model_options,
            key="model_selector",
            help="Select the flux calculation model"
        )

        #if model_name == "Dynamic system (only substrates)":
        #    st.error("Not yet implemented...")
        if model_name != "--":
            # Initialize selected model
            self.model = self.io.select_model(model_name)
            try:
                self.model.get_params()
                self.silent_sim()
            except Exception as e:
                st.error(f"The following error occurred with the selected model: {e}")
                raise
            else:
                st.success("Model successfully initialized")

            expand_basic_settings = st.expander("Basic settings",
                                                expanded=True)
            with expand_basic_settings:
                # Select model parameters
                self.mc = st.checkbox(
                    "Sensitivity analysis (Monte Carlo)",
                    value=self.defaults["mc"] if self.config_parser is None
                    else self.config_parser.mc,
                    help="Determine the precision on estimated parameters by "
                         "Monte Carlo sensitivity analysis."
                )
                enable_mc = False if self.mc else True
                self.iterations = st.number_input(
                    "Number of iterations",
                    value=self.defaults["iterations"] if self.config_parser is None
                    else self.config_parser.iterations ,
                    help="Number of iterations for Monte Carlo analysis.",
                    disabled=enable_mc
                )
                if self.iterations < 0:
                    st.error("ERROR: Number of Monte-Carlo iterations must be a positive integer")
                self.debug_mode = st.checkbox(
                    "Verbose logs",
                    help="Useful in case of trouble. Join it to the "
                         "issue on github."
                )

                if file_extension in ["tsv", "txt"]:

                    self._output_directory_selector()

                else:
                    self.io.home_path = Path(self.config_parser.path_to_data).resolve().parent
                    self.io.res_path = self.io.home_path / (self.io.home_path.name + "_res")

            # Build the form for advanced parameters
            form = st.form("Parameter_form")
            with form:
                expand_run_params = st.expander("Parameters")
                with expand_run_params:
                    st.subheader("Parameters to estimate")
                    col1, col2, col3, col4 = st.columns(4)
                    with col1:
                        st.write("Parameter Name")
                        for key in self.model.parameters_to_estimate:
                            st.text_input(
                                label="label", # Unused
                                label_visibility="collapsed",
                                value=key,
                                key=f"Parameter_name_{key}",
                                disabled=True
                            )
                    with col2:
                        st.write("Parameter Value")
                        for key, value in self.model.parameters_to_estimate.items():
                            st.text_input(
                                label="label", # Unused
                                label_visibility = "collapsed",
                                value=value if self.config_parser is None
                                else self.config_parser.model["parameters_to_estimate"][key],
                                key=f"Parameter_value_{key}"
                            )
                    with col3:
                        st.write("Lower Bound")
                        for key, bound in self.model.bounds.items():
                            st.text_input(
                                label="label",  # Unused
                                label_visibility="collapsed",
                                value=bound[0] if self.config_parser is None
                                else literal_eval(self.config_parser.model["bounds"][key])[0],
                                key=f"Parameter_lower_{key}"
                            )
                    with col4:
                        st.write("Upper Bound")
                        for key, bound in self.model.bounds.items():
                            st.text_input(
                                label="label",  # Unused
                                label_visibility="collapsed",
                                value=bound[1] if self.config_parser is None
                                else literal_eval(self.config_parser.model["bounds"][key])[1],
                                key=f"Parameter_upper_{key}"
                            )

                    if self.model.fixed_parameters is not None:
                        for param in self.model.fixed_parameters.keys():
                            st.subheader(f"Fixed parameters: {param}")
                            col1, col2 = st.columns(2)
                            with col1:
                                st.write("Parameter Name")
                                for key in self.model.fixed_parameters[param].keys():
                                    st.text_input(
                                        label="label", # Unused
                                        label_visibility="collapsed",
                                        value=key,
                                        key=f"Fixed_{param}_{key}",
                                        disabled=True
                                    )
                            with col2:
                                st.write("Parameter Value")
                                for key, value in self.model.fixed_parameters[param].items():
                                    st.text_input(
                                        label="label", # Unused
                                        label_visibility="collapsed",
                                        value=value if self.config_parser is None
                                        else self.config_parser.model["fixed_parameters"][key],
                                        key=f"Fixed_{param}_value_{key}"
                                    )

                expand_sds = st.expander("Standard Deviations")
                # Get origin of sds
                if self.config_parser is None:
                    self.sd = self.defaults["sd"]
                else:
                    self.sd = self.config_parser.sds
                with expand_sds:
                    col1, col2 = st.columns(2)
                    with col1:
                        for key in self.sd.keys():
                            st.text_input(
                                label="label",  # Unused
                                label_visibility="collapsed",
                                value=key,
                                key=f"{key}_sd",
                                disabled=True
                            )
                    with col2:
                        for key, value in self.sd.items():
                            st.text_input(
                                label="label",  # Unused
                                label_visibility="collapsed",
                                value=value if self.config_parser is None
                                else self.config_parser.sds[key],
                                key=f"Fixed_{key}_sd_value"
                            )

                submitted = st.form_submit_button("Run flux calculation")
            return submitted

    def _get_data_from_session_state(self):
        """
        Get the data from the widgets input
        """

        # Start with estimable parameters
        # Get order of parameter names to build dict
        estimable_parameter_name_order = [key for key in self.model.parameters_to_estimate.keys()]
        for name in estimable_parameter_name_order:
            try:
                # Get values from widgets
                if st.session_state[f"Parameter_value_{name}"] == "0":
                    self.model.parameters_to_estimate[name] = 0
                else:
                    self.model.parameters_to_estimate[name] = literal_eval(
                        st.session_state[f"Parameter_value_{name}"].lstrip("0") # Strip leading zeroes to stop eval errors
                    )
                # Get bounds
                if st.session_state[f"Parameter_lower_{name}"] == "0":
                    st.warning(
                        f"WARNING: {name} has a lower bound at 0. Sometimes this might confuse the optimizer. We "
                        f"strongly recommend to set the lower bound at a non-zero value, 1e-6 for example."
                    )
                    lower_bound = 0
                else:
                    lower_bound = literal_eval(st.session_state[f"Parameter_lower_{name}"].lstrip("0"))
                if st.session_state[f"Parameter_upper_{name}"] == "0":
                    st.warning(
                        f"WARNING: {name} has an upper bound at 0. Sometimes this might confuse the optimizer. We "
                        f"strongly recommend to set the lower bound at a non-zero value, 1e-6 for example."
                    )
                    upper_bound = 0
                else:
                    upper_bound = literal_eval(st.session_state[f"Parameter_upper_{name}"].lstrip("0"))
                self.model.bounds[name] = (
                    lower_bound,
                    upper_bound
                )
            except ValueError:
                st.error(
                    f"ERROR: An error occurred while parsing the input for {name}. Please check that only numbers "
                    f"have been entered."
                )
                raise

        # Do the same for each fixed parameter class
        if self.model.fixed_parameters is not None:
            fixed_parameter_classes = [param for param in self.model.fixed_parameters]
            for param in fixed_parameter_classes:
                for key in self.model.fixed_parameters[param].keys():
                    try:
                        if st.session_state[f"Fixed_{param}_value_{key}"] == "0":
                            self.model.fixed_parameters[param][key] = 0
                        else:
                            self.model.fixed_parameters[param][key] = literal_eval(
                                st.session_state[f"Fixed_{param}_value_{key}"].lstrip("0")
                            )
                    except ValueError:
                        st.error(
                            f"ERROR: An error occurred when parsing the input in the fixed parameter class {param} for "
                            f"{key}. Please check that only numbers have been entered."
                        )
                        raise

        # And finally do the same for Standard Deviations
        sd_name_order = [key for key in self.sd.keys()]
        for name in sd_name_order:
            try:
                if st.session_state[f"Fixed_{name}_sd_value"] == "0":
                    self.sd[name] = 0 # will raise Value Error as expected
                else:
                    # Get values from widgets
                    self.sd[name] = literal_eval(
                        st.session_state[f"Fixed_{name}_sd_value"].lstrip("0")  # Strip leading zeroes to stop eval errors
                    )
            except ValueError:
                st.error(
                    f"ERROR: An error occurred when parsing the input for {name} (SDs). Please check that only numbers "
                    f"have been entered and that value is superior to 0."
                )
                raise

    def _build_fitter_kwargs(self):

        kwargs = {
            "sd": self.sd,
            "model": self.model,
            "mc": self.mc,
            "iterations": self.iterations,
            "debug_mode": self.debug_mode,
        }
        return kwargs

    def _output_directory_selector(self):

        # Set up tkinter for directory chooser
        root = tk.Tk()
        root.withdraw()

        # Make folder picker dialog appear on top of other windows
        root.wm_attributes('-topmost', 1)

        # Initialize folder picker button and add logic
        clicked = st.button(
            "Select output data directory", key="clicker"
        )
        if clicked:

            # Initialize home path from directory selector and add
            # to session state
            st.session_state.home_path = Path(st.text_input(
                "Selected output data directory:",
                filedialog.askdirectory(master=root)
            ))
            if st.session_state.home_path == Path(".") \
                    or not st.session_state.home_path.exists():
                raise RuntimeError("Please provide a valid output directory")
            self.io.home_path = copy(
                st.session_state.home_path)

        elif hasattr(st.session_state, "home_path"):

            self.io.home_path = Path(st.text_input(
                "Selected output data directory:",
                st.session_state.home_path
            ))
            if self.io.home_path == Path(".") \
                    or not self.io.home_path.exists():
                raise RuntimeError("Please provide a valid output directory")

            # Initialize the result export directory
            self.io.res_path = self.io.home_path / (self.io.home_path.name + "_res")
            if not clicked:
                if not self.io.res_path.is_dir():
                    self.io.res_path.mkdir()



if __name__ == "__main__":
    physiofit_app = App()
    physiofit_app.start_app()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['physiofit',
 'physiofit.base',
 'physiofit.models',
 'physiofit.tests',
 'physiofit.ui']

package_data = \
{'': ['*'], 'physiofit': ['data/*']}

install_requires = \
['matplotlib>=3.7.1,<4.0.0',
 'numpy>=1.24.2,<2.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'pandas>=2.0.1,<3.0.0',
 'pyyaml>=6.0,<7.0',
 'scipy>=1.10.1,<2.0.0',
 'streamlit>=1.20.0,<2.0.0']

entry_points = \
{'console_scripts': ['physiofit = physiofit.__main__:main']}

setup_kwargs = {
    'name': 'physiofit',
    'version': '3.1.1',
    'description': 'Calculate extracellular fluxes from metabolite concentrations and biomass data',
    'long_description': '# PhysioFit\n\n[![PyPI version](https://badge.fury.io/py/physiofit.svg)](https://badge.fury.io/py/physiofit)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/physiofit.svg)](https://pypi.python.org/pypi/physiofit/)\n[![Documentation Status](https://readthedocs.org/projects/physiofit/badge/?version=latest)](http://physiofit.readthedocs.io/?badge=latest)\n[![install with bioconda](https://img.shields.io/badge/install%20with-bioconda-brightgreen.svg?style=flat)](http://bioconda.github.io/recipes/physiofit/README.html)\n\n\n## What is PhysioFit?\n**PhysioFit is a scientific tool designed to quantify cell growth parameters and uptake & production fluxes**\n\nFluxes are estimated using various mathematical models by fitting time-course measurements of the concentration of\ncells and extracellular substrates and products. PhysioFit v3 includes by default the most common growth models, and\nadditional models can be implemented by users.\n\nIt is one of the routine tools that we use at the [MetaSys team](http://www.lisbp.fr/en/research/molecular-physiology-and-metabolism/metasys.html) \nand [MetaToul platform](http://www.metatoul.fr) in functional studies of metabolic systems.\n\nThe code is open-source, and available under a GPLv3 license. Additional information can be found in the following \n[publication](https://doi.org/10.1128/aem.00768-19).\n\nDetailed documentation can be found online at Read the Docs \n([https://physiofit.readthedocs.io/](https://physiofit.readthedocs.io/)).\n\n## Key features\n   * **calculation of growth rate and extracellular (uptake and production) fluxes**,\n   * default models for quantifying parameters in steady-state conditions (with and without lag & metabolite degradation\n     over time),\n   * **user-defined growth models**,\n   * Monte-Carlo sensitivity analysis to **estimate the precision of the calculated fluxes**,\n   * **evaluation of the goodness of fit and visual inspection of the fitted curves**,\n   * shipped as a **library** with both a **graphical** and **command line** interface,\n   * open-source, free and easy to install everywhere where Python 3 and pip run,\n   * **biologist-friendly**.\n\n## Quick-start\nPhysioFit requires Python 3.8 or higher and run on all platforms.\nPlease check [the documentation](https://physiofit.readthedocs.io/en/latest/quickstart.html) for complete\ninstallation and usage instructions.\n\nUse `pip` to **install PhysioFit from PyPi**:\n\n```bash\n$ pip install physiofit\n```\n\nUse `conda` to install from conda:\n\n```bash\n$ conda install physiofit -c bioconda\n```\n\nThen, start the graphical interface with:\n\n```bash\n$ physiofit\n```\n\nPhysioFit is also available directly from command-line and as a Python library.\n\n## Bug and feature requests\nIf you have an idea on how we could improve PhysioFit please submit a new *issue*\nto [our GitHub issue tracker](https://github.com/MetaSys-LISBP/PhysioFit/issues).\n\n\n## Developers guide\n### Contributions\nContributions are very welcome! :heart:\n\nPlease work on your own fork,\nfollow [PEP8](https://www.python.org/dev/peps/pep-0008/) style guide,\nand make sure you pass all the tests before a pull request.\n\n### Local install with pip\nIn development mode, do a `pip install -e /path/to/PhysioFit` to install\nlocally the development version.\n\n### Build the documentation locally\nBuild the HTML documentation with:\n\n```bash\n$ cd doc\n$ make html\n```\n\nThe PDF documentation can be built locally by replacing `html` by `latexpdf`\nin the command above. You will need a recent latex installation.\n\n## How to cite\nPhysioFit: quantifying cell growth parameters and uptake and production fluxes.\nLe Grégam L., Guitton Y., Bellvert F., Jourdan F., Portais J.C., Millard P.\nIn preparation for publication\n\n## Authors\nLoïc Le Grégam, Pierre Millard\n\n## Contact\n:email: legregam@insa-toulouse.fr, millard@insa-toulouse.fr\n',
    'author': 'llegregam',
    'author_email': 'legregam@insa-toulouse.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8, !=2.7.*, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*, !=3.7.*',
}


setup(**setup_kwargs)



from __future__ import annotations

from abc import ABC, abstractmethod
from ast import literal_eval

import pandas as pd
import numpy as np


# Initialization: nom des métabolites en entrée

# Get les parametres en fonction du modèle: retourne les noms des paramètres
# (optimisables et non),leurs valeurs initiales et les bounds pour ceux qui
# sont optimisables

# Besoin d'une fonction simulate propre au modèle qui prends les temps,
# les paramètres optimisables optimisables et non optimisables


class Model(ABC):

    def __init__(self, data: pd.DataFrame):
        self.data = data
        self.time_vector = self.data.time.to_numpy()
        if "time" in self.data.columns and "experiments" in self.data.columns:
            self.name_vector = self.data.drop(["time", "experiments"], axis=1).columns.to_list()
        elif "time" in self.data.columns:
            self.name_vector = self.data.drop(["time"], axis=1).columns.to_list()
        else:
            raise ValueError(
                "Couldn't get column names from data. Is data loaded in properly and well formatted?"
            )
        self.experimental_matrix = self.data.drop("time", axis=1).to_numpy()
        self.metabolites = self.name_vector[1:]
        self.model_name = None
        self.parameters_to_estimate = None
        self.fixed_parameters = None
        self.bounds = None


    def __repr__(self):
        return f"Selected model: {self.model_name}\n" \
               f"Model data: \n{self.data}\n" \
               f"Experimental matrix: \n{self.experimental_matrix}\n" \
               f"Time vector: {self.time_vector}\n" \
               f"Name vector: {self.name_vector}\n" \
               f"Biomass & Metabolites: {self.metabolites}\n" \
               f"Parameters to estimate: {self.parameters_to_estimate}\n" \
               f"Fixed parameters: {self.fixed_parameters}\n" \
               f"Bounds: {self.bounds}\n"

    def __setattr__(self, key, value):
        if key == "data":
            self.__dict__["data"] = value
            self.time_vector = self.data.time.to_numpy()
            self.name_vector = self.data.drop("time", axis=1).columns.to_list()
            self.experimental_matrix = self.data.drop("time", axis=1).to_numpy()
            self.metabolites = self.name_vector[1:]
        else:
            self.__dict__[key] = value

    @ abstractmethod
    def get_params(self):
        """

        :return params_to_estimate: List of parameters to estimate
        :return fixed_parameters: dict of constant parameters
        :return bounds: dict of upper and lower bounds
        :return default_init_values: dict containing default initial values for
                                    params
        """
        pass

    @staticmethod
    @abstractmethod
    def simulate(
            params_opti: list,
            data_matrix: np.ndarray,
            time_vector: np.ndarray,
            params_non_opti: dict | list
    ):
        pass


class Bounds(dict):

    def __init__(self, mapping=None, **kwargs):

        if mapping is None:
            mapping = {}
        if kwargs:
            mapping.update(
                {key: value for key, value in kwargs.items()}
            )

        for key, value in mapping.items():
            key, value = self._check_bounds(key, value)
            mapping[key] = value

        super().__init__(mapping)

    def __call__(self):

        tuples = tuple(value for value in self.values())
        return tuples

    def __setitem__(self, key, value):

        key, value = self._check_bounds(key, value)
        super().__setitem__(key, value)

    @staticmethod
    def _check_bounds(key, value):

        if not isinstance(value, tuple):
            if isinstance(value, str):
                try:
                    value = literal_eval(value)
                except Exception:
                    raise TypeError(
                        f"Could not coerce {value} into string"
                    )
        for x in value:
            if not isinstance(x, int) and not isinstance(x, float):
                raise TypeError(
                    "Individual bound values must be numbers"
                )
        if value[0] >= value[1]:
            raise ValueError(
                "Lower bound cannot be higher than upper bound. "
            )
        if len(value) != 2:
            raise ValueError(
                "Bounds can only have two values. Number of values detected: "
                f"{len(value)}"
            )
        if not isinstance(key, str):
            raise TypeError(
                "Name for bounds must be strings"
            )
        return key, value


class StandardDevs(dict):

    def __init__(self, mapping=None, **kwargs):

        self._vector = None

        if mapping is None:
            mapping = {}
        if kwargs:
            mapping.update(
                {key: value for key, value in kwargs.items()}
            )

        for key, value in mapping.items():
            self._check_sd(key, value)

        super().__init__(mapping)

    @staticmethod
    def _check_sd(key, value):

        if not isinstance(key, str):
            raise TypeError(
                f"SD name field can only contain strings. Detected type {type(key)} for {key}"
            )
        if not isinstance(value, int) and not isinstance(value, float):
            try:
                value = float(value)
            except Exception:
                raise TypeError(
                    f"SD value must be a number. Detected type: {type(value)} for {key}"
                )
        if value <= 0:
            raise ValueError(
                f"SD value must be superior to 0. Detected value: {value} for {key}"
            )

    def __setitem__(self, key, value):

        self._check_sd(key, value)
        super().__setitem__(key, value)

    @property
    def vector(self):

        if self._vector is not None:
            return self._vector
        self._vector = np.array([value for value in self.values])
        return self._vector

class ModelError(Exception):
    pass


if __name__ == "__main__":
    bounds = Bounds(
        X_0=(1e-3, 10),
        mu=(1e-3, 3)
    )
    print(bounds())
